
class Rectangle:
    def __init__(self, length, width):
        self.length = length
        self.width = width

    def area(self):
        return self.length * self.width

    def perimeter(self):
        return 2 * self.length + 2 * self.width

class Square(Rectangle):
    def __init__(self, length):
        super().__init__(length, length)

class Cube(Square):
    def surface_area(self):
        face_area = super().area()
        return face_area * 6

    def volume(self):
        face_area = self.area()
        return face_area * self.length

cube = Cube(4)



import os
import numpy as np
import matplotlib.pyplot as plt
from ase import io
from ase.units import GPa
from rich.console import Console, Group
from rich.text import Text
from rich.panel import Panel
from rich.table import Table

from enstelco.deformations import Deformations
from enstelco.solve import Solvers
from enstelco.utils import get_lattice_type, colors
from enstelco.strains import STRAIN_SETS


# XXX: Adds dynamic inheritance based on lattice type
#      (I should really rework solve.py to avoid this)
def dynamic_solver(cls):
    def make_instance(atoms, **kwargs):
        if kwargs.get('lattice_type'):
            lattice_type = kwargs.get('lattice_type')
        else:
            lattice_type = get_lattice_type(atoms)

        Solver = Solvers[lattice_type]

        class RealENSTELCO(cls, Solver):
            def __init__(self):
                cls.__init__(self, atoms, **kwargs)
                Solver.__init__(self)

        return RealENSTELCO()

    return make_instance


@dynamic_solver
class ENSTELCO(Deformations):
    def __init__(self, atoms, calc=None, lattice_type=None, verbose=False,
                 input_file='POSCAR', output_file='opt.traj'):

        Deformations.__init__(
            self, atoms, calc=calc, lattice_type=lattice_type, verbose=verbose,
            input_file=input_file, output_file=output_file,
        )

    def read(self, opt_file='opt.traj'):
        n_sets = len(self.strain_set)
        strains = [np.loadtxt(f'{i:03d}/strains') for i in range(n_sets)]
        n_strains = len(strains[0])

        def get_E(path):
            return io.read(os.path.join(path, opt_file)).get_potential_energy()

        energies = [[get_E(f"{i:03d}/{j:03d}") for j in range(n_strains)]
                    for i in range(n_sets)]

        self._ref_e = self.atoms.get_potential_energy()
        self._ref_V = self.atoms.get_volume()
        self.energies = (np.array(energies) - self._ref_e) / self._ref_V
        self.strains = np.array(strains)

    def process(self):
        if self.energies is None:
            self.read()
        self.get_properties()

    def summarize(self, boring=False, file_name=None):
        if not hasattr(self, 'properties'):
            self.process()

        if file_name is not None:
            boring = True

        ec_panel = self._write_ec_panel(boring=boring)
        tensor_panel = '' if boring else self._write_tensor_panel()
        properties_panel = self._write_properties_panel(boring=boring)

        if file_name is not None:
            with open(file_name, 'w') as file:
                file.write(f'{ec_panel}{properties_panel}')
            return

        console = Console()
        display = Group(ec_panel, tensor_panel, properties_panel)
        print()
        console.print(display)
        print()

    def _write_ec_panel(self, boring=False):
        if boring:
            text = ''
            for k, v in self.elastic_constants.items():
                text += f'{k} {v}\n'
            return text

        ec_tables = []
        ec_vals = []
        for i, (k, v) in enumerate(self.elastic_constants.items()):
            if i % 7 == 0:
                ec_tables.append(Table())
                ec_vals.append([])
            ec_tables[-1].add_column(k, justify='center')
            ec_vals[-1].append(round(v, 1))

        for vals, table in zip(ec_vals, ec_tables):
            table.add_row(*list(map(str, vals)))

        ec_panel = Panel(Group(*ec_tables), title='ELASTIC CONSTANTS', expand=False)
        return ec_panel

    def _write_tensor_panel(self):
        tensor_table = Table.grid()
        for i in range(6):
            tensor_table.add_column(justify='center', min_width=8)
        for j in range(6):
            vals = np.round(self.elastic_tensor[j, :], 1)
            tensor_table.add_row(*list(map(str, vals)))

        tensor_panel = Panel(tensor_table, title='ELASTIC TENSOR', expand=False)
        return tensor_panel

    def _write_properties_panel(self, boring=False):
        if boring:
            text = ''
            def append_line(text, attr):
                text += f'{attr} {getattr(self, attr)}\n'
                return text

            attributes = [
                'K_V', 'K_R', 'K_VRH',
                'E_V', 'E_R', 'E_VRH',
                'G_V', 'G_R', 'G_VRH',
                'v_V', 'v_R', 'v_VRH',
            ]

            for a in attributes:
                text = append_line(text, a)
            return text

        headers = ['Method', 'Bulk modulus', "Young's modulus", 'Shear modulus',
                "Poisson's ratio"]
        properties_table = Table(*headers)

        mech_keys = ['Voigt', 'Reuss', 'Hill']
        for key, vals in zip(mech_keys, self.properties):
            properties_table.add_row(key, *list(map(str, np.round(vals, 2))))

        title = 'MECHANICAL PROPERTIES'
        subtitle = 'Calculated with ELATE: https://progs.coudert.name/elate'
        properties_panel = Panel(properties_table, title=title,
                                 subtitle=subtitle, expand=False)
        return properties_panel

    def plot(self, n_max=4, axs=None, save_file=None):
        etas = [i for i, _ in enumerate(STRAIN_SETS[self.lattice_type])]

        n_panels = int(np.ceil(len(etas) / n_max))
        n_cols = n_panels if n_panels < 4 else 4
        n_rows = int(np.ceil(n_panels / 4))
        width = 0.5 + 3.5 * n_cols
        height = 3 * n_rows

        if axs is None:
            fig, axs = plt.subplots(n_rows, n_cols, figsize=(width, height))
        if n_cols == n_rows == 1:
            axs = np.array([axs])
        if axs.ndim == 1:
            axs = [axs]
        k = 0
        leg_params = dict(labelcolor='linecolor', frameon=0,
                handlelength=0.5, fontsize=12)
        for a in range(n_rows):
            for b in range(n_cols):
                for n in range(n_max):
                    self.plot_eta_i(etas[k], color=colors[n], ax=axs[a][b])
                    k += 1
                    if k == len(etas):
                        break
                axs[a][b].legend(**leg_params)
                axs[a][b].tick_params(labelsize=12)
                if b == 0:
                    axs[a][b].set_ylabel('Energy', fontsize=14)
                if a + 1 == n_rows:
                    axs[a][b].set_xlabel('Strain (%)', fontsize=14)

        plt.tight_layout()
        if save_file:
            plt.savefig(save_file)
        else:
            plt.show()

    def plot_eta_i(self, i, color=None, ax=None):
        if ax is None:
            ax = plt.gca()

        data, fit_data = self._get_plot_data(i)
        text = f'$\eta$$_{i}$'
        ax.plot(*fit_data, ls='-', lw=2.0, color=color, label=text)
        ax.plot(*data, ls='', marker='o', mew=1, mec='black', color=color, ms=7)

    def _get_plot_data(self, i_read):
        A2 = self._A2[i_read]

        strains = self.strains[i_read]
        energies = self.energies[i_read] * self._ref_V + self._ref_e

        fit_strains = np.linspace(min(strains), max(strains), 30)
        fit_energies = self.get_E(A2, strains=fit_strains) * self._ref_V + self._ref_e

        return (strains * 100, energies), (fit_strains * 100, fit_energies)


if __name__ == '__main__':
    from ase.io import read
    atoms = read('opt.traj')
    enstelco = ENSTELCO(atoms)
    enstelco.read()
    enstelco.plot_cij(11)



import os
import argparse
from textwrap import dedent
from ase.io import read

from rich.console import Console
from rich.text import Text
from rich.tree import Tree
from rich.table import Table

from enstelco import __version__, ENSTELCO

commands = (
    'deform',
    'process',
    'plot',
)

class Base:
    def __init__(self, parser):
        self.parser = parser


class DeformCLI(Base):
    """
    Perform deformations on input structure with various strains to
    calculate energies and fit to elastic constants. Number of deformations
    is dependent on lattice symmetry (automatically determined or can be
    specified).

    Examples:

        $ enstelco deform CHA.cif
        $ enstelco deform --lo -0.02 --hi 0.02 thing.xyz
        $ enstelco deform -l cubic cubic_thing_but_improperly_detected_automatically.xyz
    """
    help_info = 'Deform input structure and write corresponding deformed inputs'
    def add_args(self):
        self.parser.add_argument('structure', type=str,
            help='Undeformed structure to be used for deformations'\
                    ' (any ASE-readable format)')

        self.parser.add_argument('-n', type=int, default=5,
                help='Number of strains to apply between --lo and --hi')
        self.parser.add_argument('--lo', type=float, default=0.0,
                help='Lower strain bound')
        self.parser.add_argument('--hi', type=float, default=0.04,
                help='Upper strain bound')
        self.parser.add_argument('-i', '--input_file', type=str, default='POSCAR',
                help='Name (and ext) of input file to write for each deformation')
        self.parser.add_argument('-l', '--lattice-type', type=str, default=None,
                help='Specific lattice type to force (automatically determined by'\
                        ' default). Can be spacegroup number of choose from '\
                        'enstelco.deformations.crystal_families')
        self.parser.add_argument('-s', '--silent', action='store_true',
                help='Suppress cool console output - not recommended :)')

    def main(self, args):
        atoms = read(args.structure)
        enstelco = ENSTELCO(atoms, lattice_type=args.lattice_type,
                            input_file=args.input_file)
        enstelco.deform(n=args.n, smin=args.lo, smax=args.hi)
        if not args.silent:
            self.make_tree(enstelco)

    def make_tree(self, enstelco):
        console = Console()
        text = Text(f'Writing deformations for {enstelco.lattice_type} lattice\n', 'bold')
        console.print(text)

        tree = Tree(os.getcwd())
        from rich.panel import Panel
        for d in enstelco._dirs:
            path = Text(os.path.join(d, enstelco.input_file))#, 'bold')
            tree.add(Text('📄 ') + path)
        console.print(Panel(tree, title="Deformed structures", expand=False))


class ProcessCLI(Base):
    """
    Calculate elastic constants and mechanical properties from completed
    deformation calculations.
    """
    help_info = 'Calculate elastic constants + mechanical properties'
    def add_args(self):
        self.parser.add_argument('structure', type=str,
            help='Initial undeformed structure used in calculations to initialize'\
                   ' ENSTELCO (any ASE-readable format)')
        self.parser.add_argument('-b', '--boring', action='store_true',
                help='Produce simple output for easier post-process processing')
        self.parser.add_argument('-w', '--write', metavar='file_name', default=None,
                help='Write output to file instead of displaying to console')
        self.parser.add_argument('-l', '--lattice-type', type=str, default=None,
                help='Specific lattice type to force (only use if also specified'\
                        'for deformations)')
        self.parser.add_argument('-o', '--output_file', type=str, default='opt.traj',
                help='Name (and ext) of ouput file that was written for each'\
                        ' deformation optimization')

    def main(self, args):
        atoms = read(args.structure)
        enstelco = ENSTELCO(atoms, lattice_type=args.lattice_type,
                            output_file=args.output_file)
        enstelco.process()
        enstelco.summarize(boring=args.boring, file_name=args.write)


class PlotCLI(Base):
    """
    Plot energy-strain data from completed calculations with second order fits
    for each strain tensor.

    Examples:
        $ enstelco plot start.cif
        $ enstelco plot -n 1 start.cif
        $ enstelco plot -n 1 -s energy-strain_fits.svg start.cif
    """
    help_info = 'Plot energy-strain data and fits'
    def add_args(self):
        self.parser.add_argument('structure', type=str,
            help='Initial undeformed structure used in calculations to initialize'\
                   ' ENSTELCO (any ASE-readable format)')
        self.parser.add_argument('-s', '--save', metavar='fig_name', default=None,
                help='Specify file name to save plot instead of displaying')
        self.parser.add_argument('-l', '--lattice-type', type=str, default=None,
                help='Specific lattice type to force (only use if also specified'\
                        'for deformations)')
        self.parser.add_argument('-n', type=int, default=4,
                help='Maximum number of elastic constants to plot on each'\
                        ' figure panel')
        self.parser.add_argument('-o', '--output_file', type=str, default='opt.traj',
                help='Name (and ext) of ouput file that was written for each'\
                        ' deformation optimization')

    def main(self, args):
        atoms = read(args.structure)
        enstelco = ENSTELCO(atoms, lattice_type=args.lattice_type,
                            output_file=args.output_file)
        enstelco.process()
        enstelco.plot(args.n, save_file=args.save)


cli_parsers = {
    'deform': DeformCLI,
    'process': ProcessCLI,
    'plot': PlotCLI,
}


class MyFormatter(argparse.ArgumentDefaultsHelpFormatter):
    """
    HelpFormatter that retains docstring formatting (mostly to preserve example
    listing).
    """

    def _fill_text(self, text, width, indent):
        return dedent(text)


def main():
    parser = argparse.ArgumentParser(prog='enstelco',
                                     description='ENergy-STrain ELastic COnstants'\
                                             ' and mechanical property calculations!',
                                     formatter_class=MyFormatter,
                                     )
    parser.add_argument('--version', action='version', version=__version__)
    subparsers = parser.add_subparsers(title='commands', dest='command')

    command_clis = {}
    for comm in commands:
        CLI = cli_parsers[comm]
        doc = CLI.__doc__
        subparser = subparsers.add_parser(comm, help=CLI.help_info, description=doc,
                formatter_class=MyFormatter)
        cli = CLI(subparser)
        cli.add_args()
        command_clis[comm] = cli

    parsed_args = parser.parse_args()
    command_clis[parsed_args.command].main(parsed_args)



#!/usr/bin/env python
import spglib
from numpy import array


def get_lattice_type(atoms, spacegroup=None, symprec=0.01):
    """
    finds spacegroup and corresponding crystal family (e.g., cubic)
    of ASE Atoms object

    Parameters
    ----------
    atoms
        ASE Atoms object to find lattice type

    spacegroup : int
        option to manually specify spacegroup

    symprec : float
        spglib symmetry tolerance to find spacegroup higher value for less
        symmetric structures

    Returns
    -------
    str
        lattice type of atoms
    """
    families = {
        2: "triclinic",
        15: "monoclinic",
        74: "orthorhombic",
        88: "tetragonal2",
        142: "tetragonal1",
        148: "trigonal2",
        167: "trigonal1",
        194: "hexagonal",
        230: "cubic",
    }

    if spacegroup is None:
        cell = (atoms.cell.array, atoms.get_scaled_positions(), atoms.numbers)
        s = spglib.get_spacegroup(cell, symprec=symprec)  # returns 'text (spacegroup)'
        spacegroup = int(s[s.find("(") + 1 : s.find(")")])

    for k, v in families.items():
        if spacegroup - k <= 0:
            return v


def voigt_to_full(V):
    """
    find full 3x3 array from 6 element vector in voigt notation

                                    [[xx, xy, xz],
    [xx, yy, zz, yz, xz, xy]  -->    [xy, yy, yz],
                                     [xz, yz, zz]]



    Parameters
    ----------
    V : array-like
        6 element vector in voigt notation

    Returns
    -------
    array
        3x3 array of stress tensor components
    """


    xx, yy, zz, yz, xz, xy = V
    A = array(
        [[xx, 0.5 * xy, 0.5 * xz],
         [0.5 * xy, yy, 0.5 * yz],
         [0.5 * xz, 0.5 * yz, zz]]
    )
    return A


def full_to_voigt(A):
    """
    find full (symmetric) 3x3 array from 6 element vector in voigt notation

    [[xx, xy, xz],
     [xy, yy, yz],  -->   [xx, yy, zz, yz, xz, xy]
     [xz, yz, zz]]

    Parameters
    ----------
    A : array-like
        3x3 array to transform into voigt notation vector

    Returns
    -------
    array
        6 element vector in voigt notation
    """
    A = 0.5 * (A + A.T)
    V = [A[0, 0], A[1, 1], A[2, 2], 2*A[1, 2], 2*A[0, 2], 2*A[0, 1]]
    return array(V)


#seaborn.color_palette('deep')
colors = [(0.2980392156862745, 0.4470588235294118, 0.6901960784313725),
          (0.8666666666666667, 0.5176470588235295, 0.3215686274509804),
          (0.3333333333333333, 0.6588235294117647, 0.40784313725490196),
          (0.7686274509803922, 0.3058823529411765, 0.3215686274509804),
          (0.5058823529411764, 0.4470588235294118, 0.7019607843137254),
          (0.5764705882352941, 0.47058823529411764, 0.3764705882352941),
          (0.8549019607843137, 0.5450980392156862, 0.7647058823529411),
          (0.5490196078431373, 0.5490196078431373, 0.5490196078431373),
          (0.8, 0.7254901960784313, 0.4549019607843137),
          (0.39215686274509803, 0.7098039215686275, 0.803921568627451)]


def _generate_color():
    for color in colors:
        yield color


_gen_color = _generate_color()
def next_color():
    """
    Convenience function to grab a new color during plotting loops.

    Returns
    -------
    tuple[float]
        New color rgb vals from seaborn.color_palette('deep')

    Example
    -------
        .. code-block:: python

            for values in all_values:
                plt.plot(values, color=next_color())
    """
    return next(_gen_color)



#!/usr/bin/env python
import numpy as np
import os
import glob
from ase.io import read

from enstelco.strains import STRAIN_SETS
from enstelco.utils import voigt_to_full, get_lattice_type

crystal_families = ['cubic', 'hexagonal', 'trigonal1',
                    'trigonal2', 'tetragonal1', 'tetragonal2',
                    'orthorhombic', 'monoclinic', 'triclinic']


class Deformations:
    def __init__(self, atoms, calc=None, lattice_type=None, verbose=False,
            input_file='POSCAR', output_file='opt.traj'):
        """

        Parameters
        ----------
        atoms
            relaxed ASE Atoms object to deform.

        calc
            ASE calculator object to use for energy calculations.

        lattice_type : (str or int)
            manually specify lattice_type either as str (see crystal_families)
            or int (spacegroup number).

        verbose : bool
            print various summaries and display progress if True.

        input_file : str
            name of input file to write for each deformation.

        output_file : str
            name of output file to write for each deformation calculation.
        """

        self.atoms = atoms
        self.calc = calc
        self.verbose = verbose
        self.input_file = input_file
        self.output_file = output_file

        if lattice_type is None:
            self.lattice_type = get_lattice_type(atoms)
        elif isinstance(lattice_type, int):
            self.lattice_type = get_lattice_type(atoms, spacegroup=lattice_type)
        else:
            if lattice_type.lower() not in crystal_families:
                raise ValueError(f'Lattice type not recognized, choose from:\n\
                                   {crystal_families}')
            self.lattice_type = lattice_type.lower()

        self.strain_set = STRAIN_SETS[self.lattice_type]

    def deform(self, n=5, smin=0.0, smax=0.04, i_rerun=None, strains=None):

        if strains is None:
            self.strains = np.linspace(smin, smax, n)
        else:
            self.strains = np.array(strains)

        self._dirs = []
        for i, eta  in enumerate(self.strain_set):
            if i_rerun is not None:
                if i != i_rerun:
                    continue
            eta = voigt_to_full(eta)

            for j, s in enumerate(self.strains):
                deformed_atoms = self.get_deformation(eta, s)
                path = os.path.join(f'{i:03d}', f'{j:03d}')
                os.makedirs(path, exist_ok=True)
                deformed_atoms.write(os.path.join(path, self.input_file))
                self._dirs.append(path)

            np.savetxt(os.path.join(f'{i:03d}', 'strains'), self.strains)

    def get_deformation(self, eta, strain):
        if len(eta) == 6:
            eta = voigt_to_full(eta)

        deformed_atoms = self.atoms.copy()
        A = np.identity(3) + eta * strain
        new_cell = np.dot(deformed_atoms.get_cell(), A)
        deformed_atoms.set_cell(new_cell, scale_atoms=True)
        return deformed_atoms

    def rerun(self, i_rerun, n, smin, smax, strains=None, deform=True):
        """
        Rerun specific deformation. Might be useful if certain directions give
        poor fits and you want to manually adjust strains for that direction.
        Maybe poor practice? :shrug:
        """
        os.system(f'rm -r {i_rerun:03d}/0*')
        if deform:
            self.deform(n=n, smin=smin, smax=smax, i_rerun=i_rerun, strains=strains)

        deformation_dirs = glob.glob(f'{i_rerun:03d}/0*')
        deformation_dirs.sort()

        for d in deformation_dirs:
            os.chdir(d)
            atoms = read(self.input_file)
            atoms.calc = self.calc
            atoms.get_potential_energy()
            atoms.write(self.output_file)
            os.chdir('../..')

    def run(self, n=5, smin=0.0, smax=0.04, calc=None, deform=True):
        if deform:
            self.deform(n=n, smin=smin, smax=smax)
        if calc is None:
            if self.calc is None:
                print('NO CALCULATOR PRESENT')
                print('ONLY PERFORMING DEFORMATIONS')
                return
            calc = self.calc
        if not hasattr('self', '_dirs'):
            # user might make their own dirs for whatever reason
            deformation_dirs = sorted(glob.glob('0*/0*'))
        else:
            deformation_dirs = self._dirs

        for d in deformation_dirs:
            os.chdir(d)
            atoms = read(self.input_file)
            atoms.calc = calc
            atoms.get_potential_energy()
            atoms.write(self.output_file)
            os.chdir('../..')


if __name__ == '__main__':
    from ase.io import read
    from zeoml.lmp.calculator import DeepMD
    atoms = read('opt.traj')
    graph = '/home/sours/data/univ_ml/datasets/00_zeolites/tuning/all/final/64/graph.pb'
    calc = DeepMD(graph, minimize=True)
    idk = Deformations(atoms, calc=calc)
    idk.process()



from enstelco import __version__
from setuptools import setup
from pathlib import Path

version = __version__

HERE = Path(__file__).parent

README = (HERE / 'README.md').read_text()

install_requires = (HERE / 'requirements.txt').read_text().splitlines()

python_requires = '>=3.6'

packages = ['enstelco', 'enstelco/elate']
package_data = {}

setup(name='enstelco',
      version=version,
      description='ENergy-STrain ELastic COnstant calculations made simple!',
      long_description=README,
      long_description_content_type='text/markdown',
      url='https://github.com/tysours/ENSTELCO',
      maintainer='Ty Sours',
      maintainer_email='tsours@ucdavis.edu',
      license='MIT',
      packages=packages,
      python_requires=python_requires,
      install_requires=install_requires,
      package_data=package_data,
      entry_points={'console_scripts': ['enstelco=enstelco.cli:main']},
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering :: Chemistry'],
      project_urls={
          'Source': 'https://github.com/tysours/ENSTELCO',
          'Tracker': 'https://github.com/tysours/ENSTELCO/issues',
          },
      )
